import plib


class BasePath(plib.Path):
    @property
    def yaml(self) -> dict:
        # cache results in json
        if self.json_path.mtime < self.mtime:
            self.json_path.json = super().yaml
        return self.json_path.json

    @yaml.setter
    def yaml(self, value):
        super(plib.Path, plib.Path(self)).__setattr__("yaml", value)
        self.json_path.json = value

    @property
    def json_path(self):
        return self.with_suffix(".json")


class Path(BasePath):
    assets: BasePath = BasePath.assets / "music"

    download_info = assets / "downloads"
    download_ids = download_info / "ids.yaml"
    fails = download_info / "fails.yaml"
    to_download = assets / "cache" / "to_download.yaml"
    frontend_assets = assets / "frontend"
    frontend_hash = frontend_assets / "source_code_hash.txt"

    artist_assets = assets / "artists"
    artists = artist_assets / "artists.yaml"
    recommendations = artist_assets / "recommendations.yaml"

    config = assets / "config" / "config.yaml"

    tokens = assets / "tokens" / "tokens"
    cache_assets = assets / "cache"
    cache = cache_assets / ".cache"

    download_assets: BasePath = BasePath("/") / "media" / "backup" / "Music"
    downloaded_songs = download_assets / "downloads"
    processed_songs = download_assets / "processed"
    all_songs = download_assets / "all"
    deleted = download_assets / "deleted"

    root = BasePath(__file__).parent.parent.parent
    frontend = root / "frontend"

    phone = "Music"



import os

import cli
import hostfinder
import pysftp

from music.backend.downloads import postprocessor
from music.backend.utils import Path


def start(fix_mtimes=False, port=2222):
    with cli.status("Looking for phone"):
        ip = hostfinder.find_host(port=port)
    if ip is not None:
        start_upload(ip, port=port, fix_mtimes=fix_mtimes)


def start_upload(ip, port, fix_mtimes):
    cnopts = pysftp.CnOpts()
    cnopts.hostkeys = None
    cnopts.log = True

    with cli.status("Connecting to phone"):
        sftp = pysftp.Connection(
            ip,
            port=port,
            username=os.getlogin(),
            password=cli.get("pw phone"),
            cnopts=cnopts,
        )
    with sftp:
        process_remote_deletes(sftp, fix_mtimes)
        upload(sftp)


def upload(sftp):
    sftp.makedirs(Path.phone)

    downloads = list(Path.processed_songs.glob("*.opus"))  # make list to know length
    downloads = cli.progress(
        downloads, description="Copying to phone", unit="songs", cleanup=True
    )
    for song in downloads:
        sftp.put(
            localpath=song, remotepath=f"{Path.phone}/{song.name}", preserve_mtime=True
        )
        song.rename(Path.all_songs / song.name)


def process_remote_deletes(sftp, fix_mtimes):
    with cli.console.status("Checking remote deletes"):
        phone_songs = sftp.listdir(Path.phone)

    if fix_mtimes and False:
        for path in Path.all_songs.iterdir():
            postprocessor.process_download(path, set_title=False)

    songs = Path.all_songs.iterdir()
    if fix_mtimes:
        songs = cli.progress(list(songs), description="fixing mtime", unit="songs")

    for song in songs:
        if song.name not in phone_songs:
            print(f"Removing {song.stem}")
            song.rename(Path.deleted / song.name)

        elif fix_mtimes:
            remote_path = f"{Path.phone}/{song.name}"
            mtime = sftp.stat(remote_path).st_mtime
            if int(mtime) != int(song.mtime):
                print(song.name)
                mtime = max(song.mtime, 0)
                times = (mtime, mtime)
                sftp._sftp.utime(remote_path, times)


if __name__ == "__main__":
    start()



from music.backend.artist import ArtistManager
from music.backend.artist.artists import Artist


def test_artist():
    a = Artist("Famous", "id")
    a = ArtistManager(a)
    assert isinstance(a.dict(), dict)

    albums = {"album": 1}
    a.path.albums.yaml = albums
    assert a.path.albums.yaml == albums

    a.path.albums.unlink()


test_artist()



import argparse


def get_args():
    parser = argparse.ArgumentParser()

    parser.add_argument(
        "--no-phone", dest="no_phone", action="store_true", default=False
    )
    parser.add_argument("--add", default=None)
    args = parser.parse_args()
    return args


def get_backend_args():
    parser = argparse.ArgumentParser()

    parser.add_argument("--headless", action="store_true", default=False)
    parser.add_argument("--debug", action="store_true", default=False)
    parser.add_argument("--restart", action="store_true", default=False)
    parser.add_argument(
        "--no-frontend-check",
        dest="no_frontend_check",
        action="store_true",
        default=False,
    )
    parser.add_argument("--backend", action="store_true", default=False)
    args = parser.parse_args()
    return args



from spotdl import Spotdl
from spotdl.types.options import DownloaderOptions
from spotdl.utils.config import DOWNLOADER_OPTIONS

from ..utils import Path, tokens


def get_spotdl_client():
    options = get_download_options()
    return Spotdl(
        tokens.spotify.client_id,
        tokens.spotify.client_secret,
        downloader_settings=options,
    )


def get_download_options():
    output_path_format = str(Path.downloaded_songs / "{artists} - {title}.{output-ext}")
    DOWNLOADER_OPTIONS["format"] = "opus"
    DOWNLOADER_OPTIONS["output"] = output_path_format
    DOWNLOADER_OPTIONS["threads"] = 10
    return DownloaderOptions(**DOWNLOADER_OPTIONS)



import pytest
from hypothesis import HealthCheck, given, settings, strategies

from music.backend.utils import Path


def dictionary_strategy():
    return strategies.dictionaries(keys=strategies.text(), values=strategies.text())


def nested_dictionary_generator():
    return strategies.dictionaries(
        keys=strategies.text(),
        values=dictionary_strategy(),
    )


@pytest.fixture()
def path():
    with Path.tempfile() as path:
        yield path
    path.json_path.unlink(missing_ok=True)
    for removed_path in (path, path.json_path):
        assert not removed_path.exists()


@given(content=nested_dictionary_generator())
@settings(suppress_health_check=(HealthCheck.function_scoped_fixture,), max_examples=10)
def test_yaml(path, content):
    path.yaml = content
    assert path.yaml == content


@given(content=nested_dictionary_generator())
@settings(suppress_health_check=(HealthCheck.function_scoped_fixture,), max_examples=10)
def test_json_usage(path, content):
    path.json_path.json = content
    assert path.yaml == content


@given(content1=nested_dictionary_generator(), content2=nested_dictionary_generator())
@settings(suppress_health_check=(HealthCheck.function_scoped_fixture,), max_examples=10)
def test_yaml_priority(path, content1, content2):
    path.yaml = content1
    path.json_path.json = content2
    path.json_path.mtime = path.mtime - 1
    assert path.yaml == content1



from dataclasses import dataclass

NORMAL = "normal"
FAVORITE = "favorite"


@dataclass(order=True)
class Artist:
    id: str
    name: str
    type: str = NORMAL

    @property
    def sort_index(self):
        """Favorites first and then order by name."""
        return self.type == NORMAL, self.name

    def dict(self):
        return self.__dict__

    def toggle_type(self):
        self.type = FAVORITE if self.type == NORMAL else NORMAL



from dataclasses import dataclass

import dacite


@dataclass
class Item:
    @classmethod
    def from_dict(cls, items):
        return dacite.from_dict(cls, items, config=dacite.Config(strict=True))



import random
from dataclasses import dataclass

from music.backend.artist import Artist, Artists
from music.backend.artist.artists import artists as artists_collection
from music.backend.client import response_types, spotapi
from music.backend.downloads import jobs
from music.backend.utils import Path


@dataclass
class DisplaySong(response_types.Track):
    downloaded: bool = False

    @classmethod
    def from_track(cls, track: response_types.Track):
        # don't use dict() method because it serializes all attributes recursively
        attributes = track.__dict__
        attributes["downloaded"] = jobs.is_downloaded(track)
        return DisplaySong(**attributes)


@dataclass
class DisplayArtist(response_types.ArtistInfo):
    added: bool = False

    @classmethod
    def from_artist_info(cls, artist: response_types.ArtistInfo):
        attributes = artist.__dict__
        attributes["added"] = artist.id in artists_collection.artists
        return DisplayArtist(**attributes)


def artists():
    artists = Artists()
    infos = spotapi.artists(artists.ids)
    artists = artists.artist_list()
    infos = [info.dict() | a.dict() for a, info in zip(artists, infos)]
    return infos


def add_artist(id, name):
    artists = Artists()
    artists[id] = Artist(id, name)
    artists.save()


def add_song(id):
    tracks = spotapi.songs([id])
    jobs.add(tracks)


def search_artists(name):
    return [
        DisplayArtist.from_artist_info(artist) for artist in spotapi.search_artist(name)
    ]


def search_song(name):
    return [DisplaySong.from_track(song) for song in spotapi.search_song(name)]
    # TODO: add youtube songs if spotify result list is empty


def change_artist(id: str):
    artists = Artists()
    artists[id].toggle_type()
    artists.save()


def recommendations(amount=50, max_tries=10):
    ids = Artists().ids
    random.shuffle(ids)
    seed_ids = ids[:max_tries]

    freqs = Path.recommendations.yaml
    recommendations = set({})

    while len(recommendations) < amount and seed_ids:
        id = seed_ids.pop(0)
        artists = [a for a in spotapi.related_artists(id) if a.id not in ids]
        recommendations.update(artists)
        for artist in artists:
            freqs[artist.id] = freqs.get(artist.id, 0) + 1

    Path.recommendations.yaml = freqs
    recommendations = sorted(recommendations, key=lambda r: freqs[r.id])
    return recommendations


def song_recommendations(seed_amount=5):
    downloads = Path.download_ids.yaml
    ids = list(downloads.keys())
    random.shuffle(ids)
    seed_ids = ids[:seed_amount]

    names = {v for v in downloads.values()}
    songs = [
        s
        for s in spotapi.song_recommendations(seed_ids)
        if s.id not in downloads and jobs.full_name(s) not in names
    ]
    return songs



from dataclasses import asdict, dataclass

from ..utils.item import Item as Response

dataclass = dataclass(eq=False)


@dataclass
class Urls:
    spotify: str


@dataclass
class IDS:
    isrc: str | None


@dataclass
class Item:
    external_urls: Urls
    href: str
    id: str
    name: str
    type: str
    uri: str

    def __hash__(self):
        return hash(self.id)

    def dict(self):
        return asdict(self)


@dataclass
class Followers:
    href: str | None
    total: int


@dataclass
class Image:
    height: int
    url: str
    width: int


@dataclass
class Artist(Item):
    pass


@dataclass
class ArtistInfo(Item):
    followers: Followers
    genres: list[str]
    images: list[Image]
    popularity: int


@dataclass
class Album(Item):
    album_type: str
    artists: list[Artist]
    images: list[Image]
    release_date: str
    release_date_precision: str
    total_tracks: int
    album_group: str | None
    is_playable: bool | None
    available_markets: list[str] | None
    restrictions: dict | None


@dataclass
class AlbumTrack(Item):
    artists: list[Artist]
    disc_number: int
    duration_ms: int
    explicit: bool
    is_local: bool
    is_playable: bool | None
    preview_url: str | None
    track_number: int
    linked_from: dict | None
    restrictions: dict | None


@dataclass
class Track(AlbumTrack):
    album: Album
    popularity: int
    external_ids: IDS
    available_markets: list[str] | None


@dataclass
class PaginatedResponse(Response):
    href: str
    previous: str | None
    next: str | None
    limit: int
    offset: int
    total: int


@dataclass
class ArtistSearch(PaginatedResponse):
    items: list[ArtistInfo]


@dataclass
class Artists(Response):
    artists: list[ArtistInfo]


@dataclass
class Tracks(Response):
    tracks: list[Track]


@dataclass
class Albums(PaginatedResponse):
    items: list[Album]


@dataclass
class AlbumTracks(PaginatedResponse):
    items: list[AlbumTrack]


@dataclass
class TrackSearch(PaginatedResponse):
    items: list[Track]


@dataclass
class RecommendedTracks(Tracks):
    seeds: list[dict]



import sys

import cli

from ..artist import ArtistManager, Artists
from ..client import spotapi
from ..downloads import download as start_downloads
from ..downloads import jobs
from ..utils import Path, get_args


def main():
    args = get_args()
    if args.add is not None:
        add_new_songs([args.add])

    no_phone_needed = Path.processed_songs.is_empty() or args.no_phone
    new_songs_needed = no_phone_needed and not Path.to_download.yaml
    if new_songs_needed:
        collect_new_songs()

    start_downloads()

    if "hang" in sys.argv:
        print("Press key to exit")
        input("$ ")


def collect_new_songs():
    artists = Artists()
    artists = cli.progress(
        artists, description="Checking new songs", unit="artists", cleanup=True
    )
    for artist in artists:
        ArtistManager(artist).collect_new_songs()


def add_new_songs(names):
    new_songs = []
    for name in names:
        song_results = spotapi.search_song(name)
        see_next = True

        for song in song_results:
            if see_next:
                see_next = False
                full_name = jobs.full_name(song)
                if cli.confirm(f"{full_name}\nDownload?"):
                    new_songs.append(song)
                elif cli.confirm("See next result"):
                    see_next = True

    jobs.add(new_songs)


if __name__ == "__main__":
    main()



from music.backend.utils import Path, iteration

from . import downloader, jobs, postprocessor, uploader


def download():
    songs = jobs.get()
    if songs:
        for chunk in iteration.chunked(songs, chunk_size=20):
            download_songs_chunk(chunk)
    if not Path.processed_songs.is_empty():
        uploader.start()


def download_songs_chunk(songs):
    downloader.download(songs)
    postprocessor.process_downloads()
    jobs.remove(songs)



import logging
import time

import requests
import spotipy
from retry import retry
from spotipy.cache_handler import CacheFileHandler
from spotipy.oauth2 import SpotifyClientCredentials

from music.backend.utils import Path, tokens


class Spotify(spotipy.Spotify):
    def __init__(self, market="BE"):
        self.market = market
        ccm = SpotifyClientCredentials(
            client_id=tokens.spotify.client_id,
            client_secret=tokens.spotify.client_secret,
            cache_handler=CacheFileHandler(Path.cache),
        )
        super().__init__(
            client_credentials_manager=ccm,
            requests_timeout=1,
            status_retries=0,
            retries=3,
            backoff_factor=1,
        )
        logging.disable()  # ignore retry error log messages from spotify library

    @retry(requests.exceptions.ReadTimeout, tries=10)
    def _internal_call(self, method, url, payload, params):
        try:
            for market_param in ("country", "market"):
                if market_param in params:
                    params[market_param] = self.market
            response = super()._internal_call(method, url, payload, params)
            return response
        except spotipy.exceptions.SpotifyException as e:
            if e.http_status in (404, 429):
                # include 404 status because spotify api sometimes returns 404 error
                # when it should be 429
                # https://community.spotify.com/t5/Spotify-for-Developers/
                # Intermittent-404s-Getting-Playlist-Tracks-via-API/m-p/5356770
                time.sleep(2)
                raise requests.exceptions.ReadTimeout
            else:
                raise



from __future__ import annotations

from dataclasses import dataclass

from .path import Path


@dataclass
class Config:
    backend_port: int = 13000
    hostname: str = None

    @classmethod
    def load(cls):
        return cls(**Path.config.yaml)


config = Config.load()



from ..utils import Path
from .artist import Artist

NORMAL = "normal"
FAVORITE = "favorite"


class Artists:
    def __init__(self):
        self.artists: dict[str, Artist] = {
            info["id"]: Artist(**info) for info in Path.artists.yaml
        }

    def __getitem__(self, item):
        return self.artists[item]

    def __setitem__(self, key, value):
        self.artists[key] = value

    def __iter__(self):
        yield from self.artist_list()

    def __len__(self):
        return len(self.artists)

    def artist_list(self):
        # use sort_index explicitely because dataclass ordering does not work
        return sorted(list(self.artists.values()), key=lambda a: a.sort_index)

    @property
    def ids(self):
        return [a.id for a in self.artist_list()]

    def add(self, artist):
        self[artist.id] = artist

    def save(self):
        Path.artists.yaml = [a.dict() for a in self.artist_list()]


artists = Artists()



from music.backend.client import spotapi


def test_client():
    name = "Mac Miller"

    artists = spotapi.search_artist(name)

    ids = [a.id for a in artists]
    artists = spotapi.artists(ids)

    id = artists[0].id
    artists = spotapi.related_artists(id)

    id = artists[0].id
    songs = spotapi.top_songs(id)

    ids = [s.id for s in songs]
    songs = spotapi.songs(ids)

    album_count = spotapi.album_count(id)
    albums = spotapi.albums(id, "album,single", album_count)

    songs = spotapi.album_songs(albums[0])



import cli

from music.backend.client.response_types import Track
from music.backend.utils import Path


def download_wanted(song: Track):
    skip_names = ("Interlude", "Intro", "Outro", "Live", "Instrumental")
    return (
        2 * 60 * 1000 < song.duration_ms < 10 * 60 * 1000
        and song.popularity > 15
        and not any([f" - {skip_name}" in song.name for skip_name in skip_names])
    )


def add(songs: list[Track]):
    if songs:
        songs = sorted(songs, key=lambda song: song.popularity, reverse=True)
        songs = {song.id: full_name(song) for song in songs if download_wanted(song)}

        ids = Path.download_ids.yaml
        names = {v for v in ids.values()}
        # filter based on id or name because there can be two identical songs with a
        # different name (caused by different artist) or id (caused by different album)
        new_songs = {k: v for k, v in songs.items() if k not in ids and v not in names}

        for name in new_songs.values():
            cli.console.print(name)

        Path.to_download.update(new_songs)


def is_downloaded(song: Track):
    ids = Path.download_ids.yaml
    names = {v for v in ids.values()}
    return song.id in ids or full_name(song) in names


def full_name(song: Track):
    artist_names = ", ".join([artist.name for artist in song.artists])
    name = f"{artist_names} - {song.name}"
    return name


def get():
    return list(Path.to_download.yaml.keys())


def remove(song_ids: list[str]):
    to_download = Path.to_download.yaml
    songs_dict = {s: to_download[s] for s in song_ids}
    Path.download_ids.update(songs_dict)
    Path.to_download.yaml = {
        k: v for k, v in Path.to_download.yaml.items() if k not in song_ids
    }



from music.backend.utils import Path


def main():
    Path.to_download.unlink()  # disable cache file
    todo = Path.to_download.yaml
    print(len(todo))
    downloads = [p.stem.lower() for p in Path.all_songs.iterdir()]
    download_titles = [map(d) for d in downloads]
    new_todo = {
        k: v for k, v in todo.items() if not found(v, downloads, download_titles)
    }
    Path.to_download.yaml = new_todo
    print(len(new_todo))


def found(v, downloads, download_titles):
    if v.lower() in downloads:
        return True

    if map(v) in download_titles:
        print(v)
        print(downloads[download_titles.index(map(v))])
        if input():
            return True

    return False


def map(value):
    return " - ".join(value.split(" - ")[1:]).lower()


if __name__ == "__main__":
    main()



import tbhandler
from fastapi import FastAPI, Request
from fastapi.middleware.cors import CORSMiddleware
from fastapi.responses import PlainTextResponse

from . import artistmanager

app = FastAPI()

origins = [f"http://localhost:{8080}"]
app.add_middleware(
    CORSMiddleware,
    allow_origins=origins,
    allow_credentials=True,
    allow_methods=origins,
    allow_headers=origins,
)


@app.exception_handler(Exception)
async def handler(_: Request, exc: Exception):
    tbhandler.show(exit_after=False, repeat=False)
    return PlainTextResponse(str(exc), status_code=400)


@app.get("/newartist")
async def get(name):
    return artistmanager.search_artists(name)


@app.get("/newsong")
async def get(name):
    return artistmanager.search_song(name)


@app.get("/addartist")
async def get(id, name):
    return artistmanager.add_artist(id, name)


@app.get("/addsong")
async def get(id):
    return artistmanager.add_song(id)


@app.get("/changeartist")
async def get(id):
    return artistmanager.change_artist(id)


@app.get("/recommendedartists")
async def get():
    return artistmanager.recommendations()


@app.get("/recommendedsongs")
async def get():
    return artistmanager.song_recommendations()


@app.get("/artists")
async def get():
    return artistmanager.artists()



import argparse
import os
import urllib.parse
from dataclasses import dataclass

import cli
import uvicorn

from music.backend.utils import Path, config, get_backend_args


@dataclass
class Starter:
    app_name: str = "music.backend.backend:app"
    args: argparse.Namespace = None
    session_name: str = None

    def __post_init__(self):
        self.args = get_backend_args()
        self.domain_name = urllib.parse.urlparse(config.hostname).netloc
        self.session_name = self.domain_name.split(".")[0]
        self.frontend_dist = Path("/") / "var" / "www" / self.domain_name

    def start(self):
        if self.args.backend:
            self.start_backend()
        else:
            self.start_server()

    def start_backend(self):
        reload_dirs = str(Path.root / "backend") if self.args.debug else None
        log_level = "info" if self.args.debug else "warning"
        uvicorn.run(
            self.app_name,
            port=config.backend_port,
            reload_dirs=reload_dirs,
            reload=self.args.debug,
            log_level=log_level,
        )

    def start_server(self):
        if not self.args.no_frontend_check and self.is_installed():
            self.check_frontend_distribution()

        if self.args.restart:
            cli.run(f"tmux kill-session -t {self.session_name}", check=False)
        cli.run(
            f"tmux new-session -s {self.session_name} -d musicserver --backend",
            check=False,
        )
        if not self.args.headless:
            cli.urlopen(config.hostname)

    def check_frontend_distribution(self):
        if self.frontend_dist.mtime < Path.frontend.mtime:
            content_hash = Path.frontend.content_hash
            if content_hash != Path.frontend_hash.text:
                self.generate_frontend_distribution()
                Path.frontend_hash.text = content_hash
            self.frontend_dist.mtime = Path.frontend.mtime

    def generate_frontend_distribution(self):
        self.check_frontend_dist_existence()
        if cli.return_code("which npm") != 0:
            raise Exception("This project requires npm. Please install npm first")

        commands = "npm install", "npm run build"
        cli.run_commands(*commands, cwd=Path.frontend)

        # dynamically generated paths cannot be in site-packages package directory
        if self.is_installed():
            (Path.frontend / "node_modules").rmtree()

    @classmethod
    def is_installed(cls):
        return Path.root.parent.name == "site-packages"

    def check_frontend_dist_existence(self):
        if not self.frontend_dist.exists():
            username = os.getlogin()
            cli.run("mkdir", self.frontend_dist, root=True)
            cli.run(f"chown {username}:{username}", self.frontend_dist, root=True)


def main():
    Starter().start()


if __name__ == "__main__":
    main()



from music.backend.client import spotapi
from music.backend.downloads import jobs
from music.backend.utils import Path

from .artist import FAVORITE, Artist


class ArtistPaths:
    def __init__(self, name):
        self.name = name.replace("/", "")

    @property
    def albums(self):
        path = Path.cache_assets / "albums" / self.name
        return path.with_suffix(".yaml")

    @property
    def album_count(self):
        path = Path.cache_assets / "album_counts" / self.name
        return path.with_suffix(".yaml")

    @property
    def top_songs(self):
        path = Path.cache_assets / "top_songs" / self.name
        return path.with_suffix(".yaml")


class ArtistManager(Artist):
    def __init__(self, artist):
        super().__init__(**artist.__dict__)
        self.path = ArtistPaths(self.name)

    @property
    def album_count(self):
        return self.path.album_count.yaml or 0

    def downloads(self):
        return self.path

    def collect_new_songs(self):
        if self.type == FAVORITE:
            self.check_all_songs()
        else:
            self.check_top_songs()

    def check_top_songs(self):
        cached_songs = self.path.top_songs.yaml
        songs = spotapi.top_songs(self.id)

        new_songs = [s for s in songs if s.id not in cached_songs]
        jobs.add(new_songs)

        new_songs_dict = {s.id: s.name for s in new_songs}
        self.path.top_songs.update(new_songs_dict)

    def check_all_songs(self):
        if spotapi.album_count(self.id) > self.album_count:
            albums = self.path.albums.yaml
            for album_type in ["album", "single"]:
                new_amount = spotapi.album_count(
                    self.id, album_type=album_type
                ) - albums.get(album_type, 0)

                if new_amount > 0:
                    new_albums = spotapi.albums(
                        self.id, album_type=album_type, amount=new_amount
                    )
                    for album in new_albums:
                        songs = spotapi.album_songs(album)
                        songs = spotapi.songs(
                            [song.id for song in songs]
                        )  # popularity and release_date needed
                        jobs.add(songs)
                        album_songs = {s.id: s.name for s in songs}
                        self.path.albums.update({album.id: album_songs})

                    self.path.album_count.yaml = self.album_count + len(new_albums)



import calendar
from datetime import datetime

from mutagen import oggopus

from ..utils import Path


def process_downloads():
    for download in Path.downloaded_songs.glob("*.opus"):
        if download.size == 0:
            raise Exception(f"{download} is empty file")
        else:
            process_download(download)


def process_download(download, set_title=True):
    tags = oggopus.OggOpus(download)

    title = tags["title"][0]
    time = parse_time(tags)

    if set_title:
        new_title = (
            f"{title} | {calendar.month_name[time.month][:3]} {time.day}, {time.year}"
        )
        tags["title"] = new_title
        tags.save()

    # do this after all other operations to avoid resetting mtime
    download.mtime = max(time.timestamp(), 0)
    download.rename(Path.processed_songs / download.name)


def parse_time(tags):
    date = tags["date"][0]
    parts = date.split("-")
    if len(parts) == 1:
        parts += [1, 1]

    y, m, d = parts
    y, m, d = int(y), int(m), int(d)
    return datetime(y, m, d)



from spotdl.console import download as spotdl_downloader
from spotdl.download.downloader import Downloader as SpotifyDownloader

from ..client import get_spotdl_client
from ..utils import Path

DOWNLOAD_RETRIES = 5


def clear_downloads():
    # existing downloads raise errors and can be removed because
    # successfully downloaded songs have already been moved to other folder
    Path.downloaded_songs.rmtree(remove_root=False)


def download(new_songs):
    songs = [
        id_ if "http" in id_ else f"https://open.spotify.com/track/{id_}"
        for id_ in new_songs
    ]

    clear_downloads()
    tries_left = DOWNLOAD_RETRIES
    while songs and tries_left:
        Downloader.download_songs(songs)
        songs = list(Path.downloaded_songs.glob("*.spotdlTrackingFile"))
        tries_left -= 1

    if songs and not tries_left:
        raise Exception("Max download retries reached")


class Downloader:
    downloader: SpotifyDownloader = None

    @classmethod
    def download_songs(cls, songs):
        if cls.downloader is None:
            cls.downloader = get_spotdl_client().downloader
        else:
            cls.downloader.progress_handler.song_count += len(songs)
        spotdl_downloader.download(songs, cls.downloader)



from . import _client
from .response_types import (
    Albums,
    AlbumTracks,
    Artists,
    ArtistSearch,
    RecommendedTracks,
    Tracks,
    TrackSearch,
)


def combine_chunks(f):
    """Combine function calls on chunks instead of whole array in one time
    :return: All results of the chunks.
    """

    def chunked_wrapper(*args, **kwargs):
        self, items, *args = args

        @combine_offsets
        def calculate_chunk(self, offset, limit):
            return f(self, items[offset : offset + limit], *args, **kwargs)

        return calculate_chunk(self, len(items))

    return chunked_wrapper


def combine_offsets(f, chunk_size=50):
    """Combine function calls with offset and limit instead of total amount
    :param chunk_size: Size of the chunks :return: All results of the
    chunks.
    """

    def chunked_wrapper(self, amount, *args, **kwargs):
        return [
            item
            for offset in range(0, amount, chunk_size)
            for item in f(self, *args, offset=offset, limit=chunk_size, **kwargs)
        ]

    return chunked_wrapper


class SpotApi:
    def __init__(self):
        self.api = _client.Spotify()

    def search_song(self, name):
        songs = self.api.search(name, type="track")
        return TrackSearch.from_dict(songs["tracks"]).items

    def search_artist(self, name):
        artists = self.api.search(name, type="artist")
        return ArtistSearch.from_dict(artists["artists"]).items

    @combine_chunks
    def artists(self, ids):
        artists = self.api.artists(ids)
        return Artists.from_dict(artists).artists

    def related_artists(self, id):
        artists = self.api.artist_related_artists(id)
        return Artists.from_dict(artists).artists

    def song_recommendations(self, track_ids):
        songs = self.api.recommendations(seed_tracks=track_ids)
        return RecommendedTracks.from_dict(songs).tracks

    def top_songs(self, id):
        songs = self.api.artist_top_tracks(id)
        return Tracks.from_dict(songs).tracks

    @combine_chunks
    def songs(self, ids):
        songs = self.api.tracks(ids)
        return Tracks.from_dict(songs).tracks

    def albums(self, artist_id, album_type, amount):
        return self._albums(amount, artist_id, album_type)

    @combine_offsets
    def _albums(self, artist_id, album_type, limit, offset):
        albums = self.api.artist_albums(
            artist_id, limit=limit, offset=offset, album_type=album_type
        )
        return Albums.from_dict(albums).items

    def album_songs(self, album):
        return self._album_songs(album.total_tracks, album.id)

    @combine_offsets
    def _album_songs(self, id_, offset, limit):
        songs = self.api.album_tracks(id_, limit=limit, offset=offset)
        return AlbumTracks.from_dict(songs).items

    def album_count(self, artist_id, album_type="album,single"):
        albums = self.api.artist_albums(artist_id, limit=1, album_type=album_type)
        return Albums.from_dict(albums).total



from dataclasses import dataclass

from .item import Item
from .path import Path


@dataclass
class SpotifyTokens(Item):
    client_id: str
    client_secret: str


@dataclass
class Tokens(Item):
    genius: str
    spotify: SpotifyTokens

    @classmethod
    def load(cls):
        items = Path.tokens.json
        return cls.from_dict(items)


tokens = Tokens.load()



"""Spotify changes the ids of its tracks sometimes run this file to give the
downloaded songs their most recent id.
"""

from music.backend.client import spotapi
from music.backend.utils import Path


def reverse(d):
    return {v: k for k, v in d.items()}


def main():
    downloads = Path.download_ids.yaml
    ids = list(downloads.keys())
    songs = spotapi.songs(ids)

    new_downloads = {song.id: title for song, title in zip(songs, downloads.values())}
    new_downloads = reverse(reverse(new_downloads))  # filter out duplicate song titles
    Path.download_ids.yaml = new_downloads


if __name__ == "__main__":
    main()
